'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var path = _interopDefault(require('path'));
var fs = _interopDefault(require('fs-extra'));
var execa = _interopDefault(require('execa'));

class Workspace {
  constructor(options) {
    this.options = options;
  }

  path(filePath) {
    return path.join(this.options.rootDir, filePath);
  }

  exists(filePath) {
    return fs.exists(this.path(filePath));
  }

  exec(cmd, args) {
    return execa(cmd, args, {
      stdout: this.options.stdout,
      stderr: this.options.stderr,
      cwd: this.options.rootDir
    });
  }

}

class NPM {
  constructor(options, workspace) {
    this.workspace = workspace;
  }

  static async detect(workspace) {
    if (await workspace.exists('package-lock.json')) {
      return true;
    }
  }

  install(packages) {
    return this.workspace.exec('npm', ['install', ...packages]);
  }

}

class Yarn {
  constructor(options, workspace) {
    this.workspace = workspace;
  }

  static async detect(workspace) {
    if (await workspace.exists('yarn.lock')) {
      return true;
    }
  }

  install(packages) {
    return this.workspace.exec('yarn', ['add', ...packages]);
  }

}

class Nop {
  static detect() {
    return Promise.resolve(false);
  }

  install() {
    return Promise.resolve();
  }

}

var PackageManagers = {
  yarn: Yarn,
  npm: NPM,
  nop: Nop
};

class PackageManager {
  constructor(options, workspace) {
    this.options = options;
    this.workspace = workspace;
    this.detectedPackageManager = null;
    this.packageManagers = {};
  }

  async detect() {
    // Detect PM if no preferred one is set
    if (!this.options.packageManager) {
      for (const name in PackageManagers) {
        if (await PackageManagers[name].detect(this.workspace)) {
          this.detectedPackageManager = name;
          break;
        }
      }
    } // If nothing detected


    if (!this.options.packageManager && !this.detectedPackageManager) {
      this.detectedPackageManager = 'nop'; // eslint-disable-next-line no-console

      console.warn('No package manager detected in ' + this.options.rootDir + '\n' + 'Ignoring any install command. Please use `lmify.setPackageManager(name)`');
    } // Validate it


    await this.getPackageManager();
  }

  getPackageManager() {
    const name = this.options.packageManager || this.detectedPackageManager;

    if (!this.packageManagers[name]) {
      if (!PackageManagers[name]) {
        throw new Error('Unknown package manager: ' + name);
      }

      this.packageManagers[name] = new PackageManagers[name](this.options, this.workspace);
    }

    return this.packageManagers[name];
  }

  install(packages) {
    return this.getPackageManager().install(packages);
  }

}

class LMIFY {
  constructor(options) {
    this.options = {
      stdout: process.stdout,
      stderr: process.stderr,
      rootDir: process.cwd(),
      packageManager: undefined,
      ...options
    };
    this._granters = [];
  }

  init() {
    if (!this._initPromise) {
      this._initPromise = this._init();
    }

    return this._initPromise;
  }

  async _init() {
    this.workspace = new Workspace(this.options);
    this.packageManager = new PackageManager(this.options, this.workspace);
    await this.packageManager.detect();
  }

  setPackageManager(name) {
    this.options.packageManager = name;
  }

  setRootDir(rootDir) {
    this.options.rootDir = rootDir;
    delete this._initPromise;
  }

  addGranter(granter) {
    this._granters.push(granter);
  }

  _grant(packages) {
    if (!this._granters.length) {
      return Promise.resolve(true);
    }

    return Promise.race(this._granters.map(granter => granter(packages))).then(r => Boolean(r));
  }

  async install(packages) {
    if (!packages) {
      return;
    }

    if (!Array.isArray(packages)) {
      packages = [packages];
    }

    if (!packages.length) {
      return;
    }

    const [grant] = await Promise.all([this._grant(packages), this.init()]);

    if (!grant) {
      return;
    }

    return this.packageManager.install(packages);
  }

} // Static methods

const instance = new LMIFY();
LMIFY.instance = instance;

for (const method of ['install', 'addGranter', 'setPackageManager', 'setRootDir']) {
  LMIFY[method] = LMIFY.instance[method].bind(LMIFY.instance);
}

module.exports = LMIFY;
